@@ -18,9 +18,9 @@ import collections
 import errno
 import logging
 import socket
-import threading
 
 from concurrent import futures
+from oslo_utils import timeutils
 from pymemcache import client as pymemcache_client
 import six
 
@@ -33,57 +33,6 @@ from tooz import utils
 LOG = logging.getLogger(__name__)
 
 
-class ThreadSafeClient(pymemcache_client.Client):
-    # FIXME(harlowja): get these changes upstream into pymemcache?
-    #
-    # See: https://github.com/pinterest/pymemcache/issues/34
-
-    def __init__(self, server, serializer=None, deserializer=None,
-                 connect_timeout=None, timeout=None):
-        super(ThreadSafeClient, self).__init__(server,
-                                               serializer=serializer,
-                                               deserializer=deserializer,
-                                               connect_timeout=connect_timeout,
-                                               timeout=timeout)
-        self.lock = threading.RLock()
-
-    def _store_cmd(self, name, key, expire, noreply, data, cas=None):
-        with self.lock:
-            return super(ThreadSafeClient, self)._store_cmd(name, key, expire,
-                                                            noreply, data,
-                                                            cas=cas)
-
-    def _misc_cmd(self, cmd, cmd_name, noreply):
-        with self.lock:
-            line = super(ThreadSafeClient, self)._misc_cmd(cmd, cmd_name,
-                                                           noreply)
-            return line
-
-    def _fetch_cmd(self, name, keys, expect_cas):
-        with self.lock:
-            return super(ThreadSafeClient, self)._fetch_cmd(name, keys,
-                                                            expect_cas)
-
-    def quit(self):
-        with self.lock:
-            return super(ThreadSafeClient, self).quit()
-
-    def set_many(self, values, expire=0, noreply=True):
-        with self.lock:
-            return super(ThreadSafeClient, self).set_many(values,
-                                                          expire=expire,
-                                                          noreply=noreply)
-
-    def delete_many(self, keys, noreply=True):
-        with self.lock:
-            return super(ThreadSafeClient, self).delete_many(keys,
-                                                             noreply=noreply)
-
-    def close(self):
-        with self.lock:
-            super(ThreadSafeClient, self).close()
-
-
 def _translate_failures(func):
     """Translates common pymemcache exceptions into tooz exceptions.
 
@@ -178,26 +127,46 @@ class MemcachedDriver(coordination.CoordinationDriver):
     .. _msgpack: http://msgpack.org/
     """
 
-    _GROUP_PREFIX = b'_TOOZ_GROUP_'
-    _GROUP_LEADER_PREFIX = b'_TOOZ_GROUP_LEADER_'
-    _MEMBER_PREFIX = b'_TOOZ_MEMBER_'
-    _GROUP_LIST_KEY = b'_TOOZ_GROUP_LIST'
+    #: Key prefix attached to groups (used in name-spacing keys)
+    GROUP_PREFIX = b'_TOOZ_GROUP_'
+
+    #: Key prefix attached to leaders of groups (used in name-spacing keys)
+    GROUP_LEADER_PREFIX = b'_TOOZ_GROUP_LEADER_'
+
+    #: Key prefix attached to members of groups (used in name-spacing keys)
+    MEMBER_PREFIX = b'_TOOZ_MEMBER_'
+
+    #: Key where all groups 'known' are stored.
+    GROUP_LIST_KEY = b'_TOOZ_GROUP_LIST'
+
+    #: Default socket/lock/member/leader timeout used when none is provided.
+    DEFAULT_TIMEOUT = 30
+
+    #: String used to keep a key/member alive (until it next expires).
+    STILL_ALIVE = b"It's alive!"
 
     def __init__(self, member_id, parsed_url, options):
         super(MemcachedDriver, self).__init__()
+        options = utils.collapse(options)
+        self._options = options
         self._member_id = member_id
         self._groups = set()
         self._executor = None
         self.host = (parsed_url.hostname or "localhost",
                      parsed_url.port or 11211)
-        default_timeout = options.get('timeout', ['30'])
-        self.timeout = int(default_timeout[-1])
+        default_timeout = options.get('timeout', self.DEFAULT_TIMEOUT)
+        self.timeout = int(default_timeout)
         self.membership_timeout = int(options.get(
-            'membership_timeout', default_timeout)[-1])
+            'membership_timeout', default_timeout))
         self.lock_timeout = int(options.get(
-            'lock_timeout', default_timeout)[-1])
+            'lock_timeout', default_timeout))
         self.leader_timeout = int(options.get(
-            'leader_timeout', default_timeout)[-1])
+            'leader_timeout', default_timeout))
+        max_pool_size = options.get('max_pool_size', None)
+        if max_pool_size is not None:
+            self.max_pool_size = int(max_pool_size)
+        else:
+            self.max_pool_size = None
         self._acquired_locks = []
 
     @staticmethod
@@ -212,15 +181,18 @@ class MemcachedDriver(coordination.CoordinationDriver):
             return value
         if flags == 2:
             return utils.loads(value)
-        raise Exception("Unknown serialization format '%s'" % flags)
+        raise coordination.SerializationError("Unknown serialization"
+                                              " format '%s'" % flags)
 
     @_translate_failures
     def _start(self):
-        self.client = ThreadSafeClient(self.host,
-                                       serializer=self._msgpack_serializer,
-                                       deserializer=self._msgpack_deserializer,
-                                       timeout=self.timeout,
-                                       connect_timeout=self.timeout)
+        self.client = pymemcache_client.PooledClient(
+            self.host,
+            serializer=self._msgpack_serializer,
+            deserializer=self._msgpack_deserializer,
+            timeout=self.timeout,
+            connect_timeout=self.timeout,
+            max_pool_size=self.max_pool_size)
         # Run heartbeat here because pymemcache use a lazy connection
         # method and only connect once you do an operation.
         self.heartbeat()
@@ -247,13 +219,13 @@ class MemcachedDriver(coordination.CoordinationDriver):
         self.client.close()
 
     def _encode_group_id(self, group_id):
-        return self._GROUP_PREFIX + group_id
+        return self.GROUP_PREFIX + group_id
 
     def _encode_member_id(self, member_id):
-        return self._MEMBER_PREFIX + member_id
+        return self.MEMBER_PREFIX + member_id
 
     def _encode_group_leader(self, group_id):
-        return self._GROUP_LEADER_PREFIX + group_id
+        return self.GROUP_LEADER_PREFIX + group_id
 
     @_retry.retry()
     def _add_group_to_group_list(self, group_id):
@@ -261,16 +233,16 @@ class MemcachedDriver(coordination.CoordinationDriver):
 
         :param group_id: The group id
         """
-        group_list, cas = self.client.gets(self._GROUP_LIST_KEY)
+        group_list, cas = self.client.gets(self.GROUP_LIST_KEY)
         if cas:
             group_list = set(group_list)
             group_list.add(group_id)
-            if not self.client.cas(self._GROUP_LIST_KEY,
+            if not self.client.cas(self.GROUP_LIST_KEY,
                                    list(group_list), cas):
                 # Someone updated the group list before us, try again!
                 raise _retry.Retry
         else:
-            if not self.client.add(self._GROUP_LIST_KEY,
+            if not self.client.add(self.GROUP_LIST_KEY,
                                    [group_id], noreply=False):
                 # Someone updated the group list before us, try again!
                 raise _retry.Retry
@@ -281,10 +253,10 @@ class MemcachedDriver(coordination.CoordinationDriver):
 
         :param group_id: The group id
         """
-        group_list, cas = self.client.gets(self._GROUP_LIST_KEY)
+        group_list, cas = self.client.gets(self.GROUP_LIST_KEY)
         group_list = set(group_list)
         group_list.remove(group_id)
-        if not self.client.cas(self._GROUP_LIST_KEY,
+        if not self.client.cas(self.GROUP_LIST_KEY,
                                list(group_list), cas):
             # Someone updated the group list before us, try again!
             raise _retry.Retry
@@ -304,7 +276,7 @@ class MemcachedDriver(coordination.CoordinationDriver):
 
         @_translate_failures
         def _get_groups():
-            return self.client.get(self._GROUP_LIST_KEY) or []
+            return self.client.get(self.GROUP_LIST_KEY) or []
 
         return MemcachedFutureResult(self._executor.submit(_get_groups))
 
@@ -439,7 +411,7 @@ class MemcachedDriver(coordination.CoordinationDriver):
     @_translate_failures
     def heartbeat(self):
         self.client.set(self._encode_member_id(self._member_id),
-                        "It's alive!",
+                        self.STILL_ALIVE,
                         expire=self.membership_timeout)
         # Reset the acquired locks
         for lock in self._acquired_locks:
@@ -488,13 +460,21 @@ class MemcachedDriver(coordination.CoordinationDriver):
                              self.leader_timeout)
 
     @_translate_failures
-    def run_watchers(self):
+    def run_watchers(self, timeout=None):
+        w = timeutils.StopWatch(duration=timeout)
+        w.start()
+        leftover_timeout = w.leftover(return_none=True)
+        known_groups = self.get_groups().get(timeout=leftover_timeout)
         result = []
-        for group_id in self.client.get(self._GROUP_LIST_KEY):
+        for group_id in known_groups:
+            leftover_timeout = w.leftover(return_none=True)
             try:
-                group_members = set(self._get_members(group_id))
+                group_members_fut = self.get_members(group_id)
+                group_members = group_members_fut.get(timeout=leftover_timeout)
             except coordination.GroupNotCreated:
                 group_members = set()
+            else:
+                group_members = set(group_members)
             old_group_members = self._group_members[group_id]
 
             for member_id in (group_members - old_group_members):
@@ -514,7 +494,8 @@ class MemcachedDriver(coordination.CoordinationDriver):
         for group_id, hooks in six.iteritems(self._hooks_elected_leader):
             # Try to grab the lock, if that fails, that means someone has it
             # already.
-            if self._get_leader_lock(group_id).acquire(blocking=False):
+            leader_lock = self._get_leader_lock(group_id)
+            if leader_lock.acquire(blocking=False):
                 # We got the lock
                 hooks.run(coordination.LeaderElected(
                     group_id,
