@@ -20,13 +20,9 @@ import contextlib
 from distutils import version
 import logging
 
-try:
-    from time import monotonic as _now  # noqa
-except ImportError:
-    from time import time as _now  # noqa
-
 from concurrent import futures
 from oslo_utils import strutils
+from oslo_utils import timeutils
 import redis
 from redis import exceptions
 from redis import lock as redis_locks
@@ -43,36 +39,6 @@ from tooz import utils
 LOG = logging.getLogger(__name__)
 
 
-def _version_checker(version_required):
-    """Checks server version is supported *before* running decorated method."""
-
-    if isinstance(version_required, six.string_types):
-        desired_version = version.LooseVersion(version_required)
-    elif isinstance(version_required, version.LooseVersion):
-        desired_version = version_required
-    else:
-        raise TypeError("Version decorator expects a string/version type")
-
-    def wrapper(meth):
-
-        @six.wraps(meth)
-        def decorator(self, *args, **kwargs):
-            useable, redis_version = self._check_fetch_redis_version(
-                desired_version)
-            if not useable:
-                raise tooz.NotImplemented("Redis version greater than or"
-                                          " equal to '%s' is required"
-                                          " to use this feature; '%s' is"
-                                          " being used which is not new"
-                                          " enough" % (desired_version,
-                                                       redis_version))
-            return meth(self, *args, **kwargs)
-
-        return decorator
-
-    return wrapper
-
-
 @contextlib.contextmanager
 def _translate_failures():
     """Translates common redis exceptions into tooz exceptions."""
@@ -91,15 +57,9 @@ def _translate_failures():
 class RedisLock(locking.Lock):
     def __init__(self, coord, client, name, timeout):
         self._name = "%s_%s_lock" % (coord.namespace, six.text_type(name))
-        # Avoid using lua locks to keep compatible with more versions
-        # of redis (and not just the ones that have lua support, also avoids
-        # ones that don't appear to have fully working lua support...)
-        #
-        # Issue opened: https://github.com/andymccurdy/redis-py/issues/550
-        #
-        # When that gets fixed (and detects the servers capabilities better
-        # we can likely turn this back on to being smart).
-        self._lock = redis_locks.Lock(client, self._name, timeout=timeout)
+        self._lock = redis_locks.LuaLock(client, self._name,
+                                         timeout=timeout,
+                                         thread_local=False)
         self._coord = coord
         self._acquired = False
 
@@ -190,34 +150,41 @@ class RedisDriver(coordination.CoordinationDriver):
     .. _AOF: http://redis.io/topics/persistence
     """
 
-    # The min redis version that this driver requires to operate with...
-    #
-    # NOTE(harlowja): 2.2.0 was selected since its the current version that
-    # exists on ubuntu precise, that version will work to a degree (except
-    # locks will not work with that version), in the future we can raise this
-    # as we move off of precise (and/or need newer features that the older
-    # versions of redis just don't have).
-    _MIN_VERSION = version.LooseVersion("2.2.0")
+    MIN_VERSION = version.LooseVersion("2.6.0")
+    """
+    The min redis version that this driver requires to operate with...
+    """
 
-    # Redis deletes dictionaries that have no keys in them, which means the
-    # key will disappear which means we can't tell the difference between
-    # a group not existing and a group being empty without this key being
-    # saved...
-    _GROUP_EXISTS = b'__created__'
-    _NAMESPACE_SEP = b':'
+    GROUP_EXISTS = b'__created__'
+    """
+    Redis deletes dictionaries that have no keys in them, which means the
+    key will disappear which means we can't tell the difference between
+    a group not existing and a group being empty without this key being
+    saved...
+    """
 
-    # This is for python3.x; which will behave differently when returned
-    # binary types or unicode types (redis uses binary internally it appears),
-    # so to just stick with a common way of doing this, make all the things
-    # binary (with this default encoding if one is not given and a unicode
-    # string is provided).
-    _DEFAULT_ENCODING = 'utf8'
+    #: Value used (with group exists key) to keep a group from disappearing.
+    GROUP_EXISTS_VALUE = b'1'
 
-    # These are used when extracting options from to make a client.
-    #
-    # See: http://redis-py.readthedocs.org/en/latest/ for how to use these
-    # options to configure the underlying redis client...
-    _CLIENT_ARGS = frozenset([
+    #: Default namespace for keys when none is provided.
+    DEFAULT_NAMESPACE = b'_tooz'
+
+    NAMESPACE_SEP = b':'
+    """
+    Separator that is used to combine a key with the namespace (to get
+    the **actual** key that will be used).
+    """
+
+    DEFAULT_ENCODING = 'utf8'
+    """
+    This is for python3.x; which will behave differently when returned
+    binary types or unicode types (redis uses binary internally it appears),
+    so to just stick with a common way of doing this, make all the things
+    binary (with this default encoding if one is not given and a unicode
+    string is provided).
+    """
+
+    CLIENT_ARGS = frozenset([
         'db',
         'encoding',
         'retry_on_timeout',
@@ -229,36 +196,55 @@ class RedisDriver(coordination.CoordinationDriver):
         'sentinel',
         'sentinel_fallback',
     ])
-    _CLIENT_LIST_ARGS = frozenset([
+    """
+    Keys that we allow to proxy from the coordinator configuration into the
+    redis client (used to configure the redis client internals so that
+    it works as you expect/want it to).
+
+    See: http://redis-py.readthedocs.org/en/latest/#redis.Redis
+
+    See: https://github.com/andymccurdy/redis-py/blob/2.10.3/redis/client.py
+    """
+
+    #: Client arguments that are expected/allowed to be lists.
+    CLIENT_LIST_ARGS = frozenset([
         'sentinel_fallback',
     ])
-    _CLIENT_BOOL_ARGS = frozenset([
+
+    #: Client arguments that are expected to be boolean convertible.
+    CLIENT_BOOL_ARGS = frozenset([
         'retry_on_timeout',
         'ssl',
     ])
-    _CLIENT_INT_ARGS = frozenset([
+
+    #: Client arguments that are expected to be int convertible.
+    CLIENT_INT_ARGS = frozenset([
          'db',
          'socket_keepalive',
          'socket_timeout',
     ])
-    _CLIENT_DEFAULT_SOCKET_TO = 30
+
+    #: Default socket timeout to use when none is provided.
+    CLIENT_DEFAULT_SOCKET_TO = 30
+
+    #: String used to keep a key/member alive (until it next expires).
+    STILL_ALIVE = b"Not dead!"
 
     def __init__(self, member_id, parsed_url, options):
         super(RedisDriver, self).__init__()
+        options = utils.collapse(options, exclude=self.CLIENT_LIST_ARGS)
         self._parsed_url = parsed_url
         self._options = options
-        encoding = options.get('encoding', [self._DEFAULT_ENCODING])
-        self._encoding = encoding[-1]
-        timeout = options.get('timeout', [self._CLIENT_DEFAULT_SOCKET_TO])
-        self.timeout = int(timeout[-1])
+        self._encoding = options.get('encoding', self.DEFAULT_ENCODING)
+        timeout = options.get('timeout', self.CLIENT_DEFAULT_SOCKET_TO)
+        self.timeout = int(timeout)
         self.membership_timeout = float(options.get(
-            'membership_timeout', timeout)[-1])
-        lock_timeout = options.get('lock_timeout', [self.timeout])
-        self.lock_timeout = int(lock_timeout[-1])
-        namespace = options.get('namespace', ['_tooz'])[-1]
+            'membership_timeout', timeout))
+        lock_timeout = options.get('lock_timeout', self.timeout)
+        self.lock_timeout = int(lock_timeout)
+        namespace = options.get('namespace', self.DEFAULT_NAMESPACE)
         self._namespace = self._to_binary(namespace)
         self._group_prefix = self._namespace + b"_group"
-        self._leader_prefix = self._namespace + b"_leader"
         self._beat_prefix = self._namespace + b"_beats"
         self._groups = self._namespace + b"_groups"
         self._client = None
@@ -300,9 +286,6 @@ class RedisDriver(coordination.CoordinationDriver):
     def running(self):
         return self._started
 
-    # 2.6.0 is required since internally PEXPIRE is used and that was only
-    # added in 2.6.0; so avoid using this on versions less than that...
-    @_version_checker("2.6.0")
     def get_lock(self, name):
         return RedisLock(self, self._client, name, self.lock_timeout)
 
@@ -322,25 +305,17 @@ class RedisDriver(coordination.CoordinationDriver):
             kwargs['unix_socket_path'] = parsed_url.path
         if parsed_url.password:
             kwargs['password'] = parsed_url.password
-        for a in cls._CLIENT_ARGS:
+        for a in cls.CLIENT_ARGS:
             if a not in options:
                 continue
-            # The reason the last index is used is that when multiple options
-            # of the same name are given via a url the values will be
-            # accumulated in a list (and not just be a single value)...
-            #
-            # For ex: the following is a valid url which will have 2 values
-            # for the 'timeout' argument:
-            #
-            # redis://localhost:6379?timeout=5&timeout=2
-            if a in cls._CLIENT_BOOL_ARGS:
-                v = strutils.bool_from_string(options[a][-1])
-            elif a in cls._CLIENT_LIST_ARGS:
+            if a in cls.CLIENT_BOOL_ARGS:
+                v = strutils.bool_from_string(options[a])
+            elif a in cls.CLIENT_LIST_ARGS:
                 v = options[a]
-            elif a in cls._CLIENT_INT_ARGS:
-                v = int(options[a][-1])
+            elif a in cls.CLIENT_INT_ARGS:
+                v = int(options[a])
             else:
-                v = options[a][-1]
+                v = options[a]
             kwargs[a] = v
         if 'socket_timeout' not in kwargs:
             kwargs['socket_timeout'] = default_socket_timeout
@@ -385,35 +360,39 @@ class RedisDriver(coordination.CoordinationDriver):
             # to so that the basic set of features we support will actually
             # work (instead of blowing up).
             new_enough, redis_version = self._check_fetch_redis_version(
-                self._MIN_VERSION)
+                self.MIN_VERSION)
             if not new_enough:
                 raise tooz.NotImplemented("Redis version greater than or"
                                           " equal to '%s' is required"
                                           " to use this driver; '%s' is"
                                           " being used which is not new"
-                                          " enough" % (self._MIN_VERSION,
+                                          " enough" % (self.MIN_VERSION,
                                                        redis_version))
             self.heartbeat()
             self._started = True
 
     def _encode_beat_id(self, member_id):
-        return self._NAMESPACE_SEP.join([self._beat_prefix,
-                                         self._to_binary(member_id)])
+        return self.NAMESPACE_SEP.join([self._beat_prefix,
+                                        self._to_binary(member_id)])
 
     def _encode_member_id(self, member_id):
         member_id = self._to_binary(member_id)
-        if member_id == self._GROUP_EXISTS:
+        if member_id == self.GROUP_EXISTS:
             raise ValueError("Not allowed to use private keys as a member id")
         return member_id
 
     def _decode_member_id(self, member_id):
         return self._to_binary(member_id)
 
+    def _encode_group_leader(self, group_id):
+        group_id = self._to_binary(group_id)
+        return b"leader_of_" + group_id
+
     def _encode_group_id(self, group_id, apply_namespace=True):
         group_id = self._to_binary(group_id)
         if not apply_namespace:
             return group_id
-        return self._NAMESPACE_SEP.join([self._group_prefix, group_id])
+        return self.NAMESPACE_SEP.join([self._group_prefix, group_id])
 
     def _decode_group_id(self, group_id):
         return self._to_binary(group_id)
@@ -421,20 +400,10 @@ class RedisDriver(coordination.CoordinationDriver):
     def heartbeat(self):
         with _translate_failures():
             beat_id = self._encode_beat_id(self._member_id)
-            # Use milliseconds if we can (which are more accurate than
-            # just seconds); but this PSETEX support was added in 2.6.0 or
-            # newer so we can only use it then...
-            supports_psetex, _version = self._check_fetch_redis_version(
-                '2.6.0', not_existent=False)
-            if not supports_psetex:
-                expiry_secs = max(0, int(self.membership_timeout))
-                self._client.setex(beat_id, time=expiry_secs,
-                                   value=b"Not dead!")
-            else:
-                expiry_ms = max(0, int(self.membership_timeout * 1000.0))
-                self._client.psetex(beat_id, time_ms=expiry_ms,
-                                    value=b"Not dead!")
-        for lock in self._acquired_locks:
+            expiry_ms = max(0, int(self.membership_timeout * 1000.0))
+            self._client.psetex(beat_id, time_ms=expiry_ms,
+                                value=self.STILL_ALIVE)
+        for lock in self._acquired_locks.copy():
             try:
                 lock.heartbeat()
             except coordination.ToozError:
@@ -491,11 +460,18 @@ class RedisDriver(coordination.CoordinationDriver):
         def _create_group(p):
             if p.exists(encoded_group):
                 raise coordination.GroupAlreadyExist(group_id)
+            # Start the transaction, all following should no longer be
+            # doing any reads (and should typically just be mutations from
+            # here on out). They need to be idempotent since transactions
+            # can fail half-way through if they key is of the wrong type
+            # or something else happens ... (redis doesn't have transactions
+            # that support rollbacks).
+            p.multi()
             p.sadd(self._groups,
                    self._encode_group_id(group_id, apply_namespace=False))
             # Add our special key to avoid redis from deleting the dictionary
             # when it becomes empty (which is not what we currently want)...
-            p.hset(encoded_group, self._GROUP_EXISTS, '1')
+            p.hset(encoded_group, self.GROUP_EXISTS, self.GROUP_EXISTS_VALUE)
 
         return RedisFutureResult(self._submit(self._client.transaction,
                                               _create_group, encoded_group,
@@ -512,6 +488,7 @@ class RedisDriver(coordination.CoordinationDriver):
             if not p.hexists(encoded_group, encoded_member_id):
                 raise coordination.MemberNotJoined(group_id, self._member_id)
             else:
+                p.multi()
                 p.hset(encoded_group, encoded_member_id,
                        self._dumps(capabilities))
 
@@ -527,7 +504,9 @@ class RedisDriver(coordination.CoordinationDriver):
         def _leave_group(p):
             if not p.exists(encoded_group):
                 raise coordination.GroupNotCreated(group_id)
-            c = p.hdel(encoded_group, encoded_member_id)
+            p.multi()
+            p.hdel(encoded_group, encoded_member_id)
+            c = p.execute()[0]
             if c == 0:
                 raise coordination.MemberNotJoined(group_id, self._member_id)
             else:
@@ -546,7 +525,7 @@ class RedisDriver(coordination.CoordinationDriver):
             potential_members = []
             for m in p.hkeys(encoded_group):
                 m = self._decode_member_id(m)
-                if m != self._GROUP_EXISTS:
+                if m != self.GROUP_EXISTS:
                     potential_members.append(m)
             if not potential_members:
                 return []
@@ -565,16 +544,13 @@ class RedisDriver(coordination.CoordinationDriver):
                     gone_members.add(potential_member)
             # Trash all the members that no longer are with us... RIP...
             if gone_members:
-                many_at_once, _version = self._check_fetch_redis_version(
-                    "2.4.0", not_existent=False)
-                if not many_at_once:
-                    for m in gone_members:
-                        p.hdel(encoded_group, self._encode_member_id(m))
-                else:
-                    encoded_gone_members = [self._encode_member_id(m)
-                                            for m in gone_members]
-                    p.hdel(encoded_group, *encoded_gone_members)
-                return [m for m in potential_members if m not in gone_members]
+                p.multi()
+                encoded_gone_members = list(self._encode_member_id(m)
+                                            for m in gone_members)
+                p.hdel(encoded_group, *encoded_gone_members)
+                p.execute()
+                return list(m for m in potential_members
+                            if m not in gone_members)
             else:
                 return potential_members
 
@@ -606,8 +582,10 @@ class RedisDriver(coordination.CoordinationDriver):
         def _join_group(p):
             if not p.exists(encoded_group):
                 raise coordination.GroupNotCreated(group_id)
-            c = p.hset(encoded_group, encoded_member_id,
-                       self._dumps(capabilities))
+            p.multi()
+            p.hset(encoded_group, encoded_member_id,
+                   self._dumps(capabilities))
+            c = p.execute()[0]
             if c == 0:
                 # Field already exists...
                 raise coordination.MemberAlreadyExist(group_id,
@@ -624,15 +602,18 @@ class RedisDriver(coordination.CoordinationDriver):
         encoded_group = self._encode_group_id(group_id)
 
         def _delete_group(p):
-            # An empty group still have the special key _GROUP_EXISTS set, so
+            # An empty group still have the special key GROUP_EXISTS set, so
             # its len is 1
             if p.hlen(encoded_group) > 1:
                 raise coordination.GroupNotEmpty(group_id)
-            if not p.delete(encoded_group):
-                raise coordination.GroupNotCreated(group_id)
+            p.multi()
+            p.delete(encoded_group)
             p.srem(self._groups,
                    self._encode_group_id(group_id,
                                          apply_namespace=False))
+            was_group_deleted, was_group_member_removed = p.execute()
+            if not any([was_group_deleted, was_group_member_removed]):
+                raise coordination.GroupNotCreated(group_id)
 
         return RedisFutureResult(self._submit(self._client.transaction,
                                               _delete_group,
@@ -671,26 +652,39 @@ class RedisDriver(coordination.CoordinationDriver):
     def unwatch_leave_group(self, group_id, callback):
         return super(RedisDriver, self).unwatch_leave_group(group_id, callback)
 
-    @staticmethod
-    def watch_elected_as_leader(group_id, callback):
-        raise tooz.NotImplemented
+    def watch_elected_as_leader(self, group_id, callback):
+        return super(RedisDriver, self).watch_elected_as_leader(
+            group_id, callback)
 
-    @staticmethod
-    def unwatch_elected_as_leader(group_id, callback):
-        raise tooz.NotImplemented
+    def unwatch_elected_as_leader(self, group_id, callback):
+        return super(RedisDriver, self).unwatch_elected_as_leader(
+            group_id, callback)
+
+    def _get_leader_lock(self, group_id):
+        name = self._encode_group_leader(group_id)
+        return self.get_lock(name)
+
+    def _run_leadership(self, watch):
+        for group_id, hooks in six.iteritems(self._hooks_elected_leader):
+            if watch.expired():
+                return
+            leader_lock = self._get_leader_lock(group_id)
+            if leader_lock.acquire(blocking=False):
+                # We got the lock
+                hooks.run(coordination.LeaderElected(group_id,
+                                                     self._member_id))
 
     def run_watchers(self, timeout=None):
-        if timeout is not None:
-            started_at = _now()
+        w = timeutils.StopWatch(duration=timeout)
+        w.start()
         result = []
-        for group_id in self.get_groups().get(timeout=timeout):
-            leftover_timeout = None
-            if timeout is not None:
-                elapsed = max(0.0, _now() - started_at)
-                leftover_timeout = timeout - elapsed
+        leftover_timeout = w.leftover(return_none=True)
+        known_groups = self.get_groups().get(timeout=leftover_timeout)
+        for group_id in known_groups:
+            leftover_timeout = w.leftover(return_none=True)
             try:
-                group_members = self.get_members(group_id).get(
-                    timeout=leftover_timeout)
+                group_members_fut = self.get_members(group_id)
+                group_members = group_members_fut.get(timeout=leftover_timeout)
             except coordination.GroupNotCreated:
                 group_members = set()
             else:
@@ -716,6 +710,7 @@ class RedisDriver(coordination.CoordinationDriver):
                         coordination.MemberLeftGroup(group_id,
                                                      member_id)))
             self._group_members[group_id] = group_members
+        self._run_leadership(w)
         return result
 
 
